@@ -1,7 +1,4 @@
 /***************************************************************************
- $RCSfile$
-                             -------------------
-    cvs         : $Id: dtaus_p.h 1104 2007-01-09 16:55:35Z christian $
     begin       : Mon Mar 01 2004
     copyright   : (C) 2004 by Martin Preuss
     email       : martin@libchipcard.de
@@ -26,16 +23,37 @@ struct AH_IMEXPORTER_DTAUS {
 };
 
 
-AB_IMEXPORTER* dtaus_factory(AB_BANKING *ab, GWEN_DB_NODE *db);
+AQBANKING_EXPORT
+GWEN_PLUGIN *imexporters_dtaus_factory(GWEN_PLUGIN_MANAGER *pm,
+				       const char *name,
+				       const char *fileName);
+
+static
+AB_IMEXPORTER *AB_Plugin_ImExporterDTAUS_Factory(GWEN_PLUGIN *pl,
+						 AB_BANKING *ab,
+						 GWEN_DB_NODE *db);
+
+static
 void GWENHYWFAR_CB AH_ImExporterDTAUS_FreeData(void *bp, void *p);
 
+static
 int AH_ImExporterDTAUS_Import(AB_IMEXPORTER *ie,
                               AB_IMEXPORTER_CONTEXT *ctx,
-                              GWEN_BUFFEREDIO *bio,
-                              GWEN_DB_NODE *params);
+                              GWEN_IO_LAYER *io,
+			      GWEN_DB_NODE *params,
+			      uint32_t guiid);
 
-int AH_ImExporterDTAUS_CheckFile(AB_IMEXPORTER *ie, const char *fname);
+static
+int AH_ImExporterDTAUS_Export(AB_IMEXPORTER *ie,
+			      AB_IMEXPORTER_CONTEXT *ctx,
+			      GWEN_IO_LAYER *io,
+			      GWEN_DB_NODE *params,
+			      uint32_t guiid);
 
+static
+int AH_ImExporterDTAUS_CheckFile(AB_IMEXPORTER *ie, const char *fname, uint32_t guiid);
+
+static
 int AH_ImExporterDTAUS__ImportFromGroup(AB_IMEXPORTER_CONTEXT *ctx,
                                         GWEN_DB_NODE *db,
                                         GWEN_DB_NODE *dbParams);
